use std::{collections::HashMap, path::PathBuf, str::FromStr};

use alloy::primitives::U256;
use alloy::primitives::{aliases::U112, Address};

use itertools::Itertools;
use miette::{miette, Result};
use serde::de::{self, Visitor};
use serde::{Deserialize, Deserializer, Serialize};

use log::{debug, info};

#[derive(Debug, Serialize, Deserialize)]
pub struct Pair {
    pub token0: Address,
    pub token1: Address,
    pub reserve0: U112,
    pub reserve1: U112,
    pub factory: Address,
}

#[derive(Debug, Eq, Hash, PartialEq)]
struct AddressPair(Address, Address);

impl Serialize for AddressPair {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        serializer.serialize_str(format!("{:?}:{:?}", self.0, self.1).as_str())
    }
}

impl<'de> Deserialize<'de> for AddressPair {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        struct AddressPairVisitor;

        impl Visitor<'_> for AddressPairVisitor {
            type Value = AddressPair;

            fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
                formatter.write_str("a string containing two addresses separated by a colon (e.g. \"0x...0:0x...0\")")
            }

            fn visit_str<E>(self, value: &str) -> Result<Self::Value, E>
            where
                E: de::Error,
            {
                let parts: Vec<&str> = value.splitn(2, ':').map(str::trim).collect();
                if parts.len() != 2 {
                    return Err(E::custom(format!(
                        "Invalid format, expected 'address1:address2' but got \"{}\"",
                        value
                    )));
                }

                let token0 = Address::from_str(parts[0])
                    .map_err(|e| E::custom(format!("Error parsing the address \"{}\"", e)))?;
                let token1 = Address::from_str(parts[1])
                    .map_err(|e| E::custom(format!("Error parsing the address \"{}\"", e)))?;

                Ok(AddressPair(token0, token1))
            }
        }

        deserializer.deserialize_str(AddressPairVisitor)
    }
}

#[derive(Debug)]
pub struct ArbitrageOpportunity {
    pair_a: Address,
    pair_b: Address,
    direction: bool, // true means token0 -> token1 -> token0
    optimum: U256,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Pairs {
    pairs: HashMap<Address, Pair>,
    by_tokens: HashMap<AddressPair, Vec<Address>>,
}

impl Pairs {
    pub fn new(filename: &PathBuf) -> Result<Self> {
        debug!("Looking for pairs file in {:?}", filename);

        let path = std::path::Path::new(&filename);
        if path.exists() {
            info!("Found existing {:?}", filename);

            let data = std::fs::read_to_string(path).map_err(|e| miette!(e))?;
            serde_json::from_str(&data).map_err(|e| miette!(e))
        } else {
            info!(
                "File {:?} doesn't exist, creating new structure from scratch",
                filename
            );

            Ok(Pairs {
                pairs: HashMap::new(),
                by_tokens: HashMap::new(),
            })
        }
    }

    #[allow(dead_code)]
    pub fn get(&self, address: Address) -> Option<&Pair> {
        self.pairs.get(&address)
    }

    pub fn get_tokens(&self, address: Address) -> Option<(Address, Address)> {
        self.pairs
            .get(&address)
            .map(|pair| (pair.token0, pair.token1))
    }

    pub fn add(
        &mut self,
        pair: Address,
        token0: Address,
        token1: Address,
        reserve0: U112,
        reserve1: U112,
        factory: Address,
    ) -> Option<()> {
        let old = self.pairs.insert(
            pair,
            Pair {
                token0,
                token1,
                reserve0,
                reserve1,
                factory,
            },
        );

        match old {
            Some(old) => {
                info!("Pair {} already present, updating reserves (reserve0: {} -> {}; reserve1: {} -> {})",
                    pair, old.reserve0, reserve0, old.reserve1, reserve1
                );

                Some(())
            }
            None => {
                info!("First time seeing pair {}, adding it", { pair });

                match self.by_tokens.get_mut(&AddressPair(token0, token1)) {
                    Some(pairs) => {
                        pairs.push(pair);
                        info!(
                            "Already know {} pairs with tokens {:?} and {:?}",
                            pairs.len(),
                            token0,
                            token1
                        );
                    }
                    None => {
                        self.by_tokens
                            .insert(AddressPair(token0, token1), vec![pair]);
                        info!(
                            "It's the first pair with tokens {:?} and {:?}",
                            token0, token1
                        );
                    }
                }

                None
            }
        }
    }

    pub fn dump(&self, filename: &PathBuf) -> Result<()> {
        let data = serde_json::to_string(&self).map_err(|e| miette!(e))?;
        std::fs::write(filename, data).map_err(|e| miette!(e))?;

        info!("{} Pairs saved to {:?}", self.pairs.len(), filename);

        Ok(())
    }

    pub fn len(&self) -> usize {
        self.pairs.len()
    }

    pub fn iter(&self) -> std::collections::hash_map::Iter<'_, Address, Pair> {
        self.pairs.iter()
    }

    pub fn get_addresses(&self) -> Vec<Address> {
        self.pairs.keys().cloned().collect()
    }

    pub fn get_reserves(&self, address: Address) -> Option<(U112, U112)> {
        self.pairs
            .get(&address)
            .map(|pair| (pair.reserve0, pair.reserve1))
    }

    pub fn update_reserves(
        &mut self,
        address: Address,
        reserve0: U112,
        reserve1: U112,
    ) -> Result<()> {
        if let Some(pair) = self.pairs.get_mut(&address) {
            pair.reserve0 = reserve0;
            pair.reserve1 = reserve1;
            info!(
                "Updated reserves for pair {}: reserve0: {}, reserve1: {}",
                address, reserve0, reserve1
            );
            Ok(())
        } else {
            debug!("Pair {} not found", address);
            Ok(()) // TODO return Err
        }
    }

    // TODO at the moment we return all the opportunities, instead we should return only the two opportunities
    // (token0 -> token1 -> token0 and token1 -> token0 -> token1) with the highest amountIn
    // Remember: choosing an opportunity invalidates the other ones
    pub fn look_for_opportunity(
        &self,
        token0: Address,
        token1: Address,
    ) -> Vec<ArbitrageOpportunity> {
        let mut opportunities: Vec<ArbitrageOpportunity> = Vec::new();
        if let Some(pairs) = self.by_tokens.get(&AddressPair(token0, token1)) {
            pairs.iter()
                .permutations(2)
                .any(|pairs| {
                    let pair_a = self.get(*pairs[0]).unwrap();
                    let pair_b = self.get(*pairs[1]).unwrap();

                    if let Some(optimum) = optimal_in(pair_a.reserve0, pair_a.reserve1, pair_b.reserve0, pair_b.reserve1) {
                        info!("Found arbitrage opportunity between pairs {} and {} swapping {} along token0 -> token1 -> token0", pairs[0], pairs[1], optimum);
                        opportunities.push(ArbitrageOpportunity{
                            pair_a: *pairs[0],
                            pair_b: *pairs[1],
                            direction: true,
                            optimum
                        });
                    }
                    if let Some(optimum) = optimal_in(pair_a.reserve1, pair_a.reserve0, pair_b.reserve1, pair_b.reserve0) {
                        info!("Found arbitrage opportunity between pairs {} and {} swapping {} along token1 -> token0 -> token1", pairs[0], pairs[1], optimum);
                        opportunities.push(ArbitrageOpportunity{
                            pair_a: *pairs[0],
                            pair_b: *pairs[1],
                            direction: false,
                            optimum
                        });
                    }
                    false
                });
        }
        opportunities
    }
}

fn optimal_in(x_a: U112, y_a: U112, x_b: U112, y_b: U112) -> Option<U256> {
    let x_a = U256::from(x_a);
    let x_b = U256::from(x_b);
    let y_a = U256::from(y_a);
    let y_b = U256::from(y_b);
    let f = U256::from(997);
    let ff = f.pow(U256::from(2));
    let _1000 = U256::from(1000);
    let _1000000 = U256::from(1000000);

    let k = f * x_b / _1000 + ff / _1000 * x_a / _1000;
    let phi = (ff * x_a * y_b * x_b * y_a / _1000000).root(2);
    let psi = y_a * x_b;

    if psi >= phi {
        None
    } else {
        Some((phi - psi) / k)
    }
}
