#![allow(clippy::too_many_arguments)]

use crate::config::Config;
use alloy::{
    eips::BlockNumberOrTag,
    primitives::{aliases::U112, Address, U256},
    providers::{
        fillers::{BlobGasFiller, ChainIdFiller, FillProvider, GasFiller, JoinFill, NonceFiller},
        Provider, ProviderBuilder, RootProvider, WsConnect,
    },
    rpc::{client::RpcClient, types::Filter},
    transports::layers::RetryBackoffLayer,
};
use futures_util::StreamExt;
use log::{debug, info};
use miette::miette;
use std::{
    collections::HashSet,
    sync::{Arc, Mutex},
    time::{Duration, Instant},
};
#[path = "pairs.rs"]
mod pairs;
use pairs::Pairs;
use tokio::sync::mpsc;

alloy::sol!(
    #[allow(missing_docs)]
    #[sol(rpc)]
    IUniswapV2Pair,
    "abi/IUniswapV2Pair.json"
);

#[derive(Debug)]
pub enum Action {
    ProcessNewPair(Address),
    ProcessOldPair(Address, U256, U256, U256, U256),
}

type AlloyProvider = FillProvider<
    JoinFill<
        alloy::providers::Identity,
        JoinFill<GasFiller, JoinFill<BlobGasFiller, JoinFill<NonceFiller, ChainIdFiller>>>,
    >,
    RootProvider,
>;

async fn subscribe(
    provider: Arc<AlloyProvider>,
    pairs: Arc<Mutex<Pairs>>,
    tx: mpsc::Sender<Action>,
) -> eyre::Result<()> {
    let filter = Filter::new()
        .event("Swap(address,uint256,uint256,uint256,uint256,address)") // TODO manage also sync and skim
        .from_block(BlockNumberOrTag::Latest);

    let sub = provider.subscribe_logs(&filter).await?;
    let mut stream = sub.into_stream();

    let mut latest_block = 0;

    while let Some(log) = stream.next().await {
        let block_number = log.block_number.unwrap();
        if block_number > latest_block {
            latest_block = block_number;
            info!("Processing block number {:?}", block_number);
        }

        let IUniswapV2Pair::Swap {
            sender: _,
            amount0In,
            amount1In,
            amount0Out,
            amount1Out,
            to: _,
        } = log.log_decode()?.inner.data;

        let pair_address = log.address();
        let pair_already_known = pairs.lock().unwrap().get(pair_address).is_some();

        debug!("Event by pair {:?}", pair_address);

        if pair_already_known {
            tx.send(Action::ProcessOldPair(
                pair_address,
                amount0In,
                amount1In,
                amount0Out,
                amount1Out,
            ))
            .await?;
        } else {
            tx.send(Action::ProcessNewPair(pair_address)).await?;
        }
    }

    Ok(())
}

async fn process_new_pair(
    pairs: Arc<Mutex<Pairs>>,
    pair_address: Address,
    provider: Arc<AlloyProvider>,
) -> eyre::Result<()> {
    let result: eyre::Result<()> = async {
        let pair = IUniswapV2Pair::new(pair_address, provider.clone()); // todo can avoid che clone?
        let token0 = pair.token0().call().await?._0;
        let token1 = pair.token1().call().await?._0;
        let reserve0 = pair.getReserves().call().await?.reserve0;
        let reserve1 = pair.getReserves().call().await?.reserve1;
        let factory = pair.factory().call().await?._0;

        pairs
            .lock()
            .unwrap()
            .add(pair_address, token0, token1, reserve0, reserve1, factory);

        Ok(())
    }
    .await;

    if let Err(e) = &result {
        eprintln!("error adding the new pair {}: {}", pair_address, e);
    }

    result
}

fn process_old_pair(
    pairs: Arc<Mutex<Pairs>>,
    pair_address: Address,
    amount0_in: U256,
    amount1_in: U256,
    amount0_out: U256,
    amount1_out: U256,
) -> eyre::Result<()> {
    let (reserve0, reserve1) = pairs.lock().unwrap().get_reserves(pair_address).unwrap();
    pairs
        .lock()
        .unwrap()
        .update_reserves(
            pair_address,
            reserve0 - U112::from(amount0_out) + U112::from(amount0_in),
            reserve1 - U112::from(amount1_in) + U112::from(amount1_out),
        )
        .unwrap(); // TODO manage error

    Ok(()) // TODO manage errors
}

async fn process_known_pairs(
    pairs: Arc<Mutex<Pairs>>,
    provider: Arc<AlloyProvider>,
) -> eyre::Result<()> {
    let addresses = pairs.lock().unwrap().get_addresses();
    let len = addresses.len();

    info!("Recovering state of {:?} saved pairs", len);

    for (i, address) in addresses.into_iter().enumerate() {
        info!("Processing pair {}/{}: {:?}", i + 1, len, address);

        let result: eyre::Result<()> = async {
            let pair = IUniswapV2Pair::new(address, provider.clone());
            let reserves = pair.getReserves().call().await?;
            let reserve0 = reserves.reserve0;
            let reserve1 = reserves.reserve1;

            let _ = pairs
                .lock()
                .unwrap()
                .update_reserves(address, reserve0, reserve1); // TODO manage error, should be ok however

            Ok(())
        }
        .await;

        if let Err(e) = &result {
            eprintln!("Error processing pair {}: {}", address, e);
            return result;
        }
    }
    Ok(())
}

fn look_for_opportunity(pairs: Arc<Mutex<Pairs>>, involved_pairs: &HashSet<Address>) {
    let pairs = pairs.lock().unwrap();

    for pair_address in involved_pairs {
        let (token0, token1) = pairs.get_tokens(*pair_address).unwrap();
        let _opportunities = pairs.look_for_opportunity(token0, token1);
    }
}

async fn manage_interruption(pairs: Arc<Mutex<Pairs>>, config: Config) -> eyre::Result<()> {
    tokio::signal::ctrl_c()
        .await
        .expect("Error receiving the signal");
    info!("Program correctly interrupted by the user");

    match pairs.lock().unwrap().dump(&config.pairs_file) {
        Ok(_) => info!("Pairs correctly dumped to {:?}", &config.pairs_file),
        Err(e) => info!("Error dumping pairs to {:?} {:?}", &config.pairs_file, e),
    }

    std::process::exit(0);
}

pub fn run(config: Config) -> miette::Result<()> {
    let runtime = tokio::runtime::Runtime::new().map_err(|e| miette!(e))?;
    let pairs = Arc::new(Mutex::new(Pairs::new(&config.pairs_file)?));
    let ws = WsConnect::new(&config.endpoint);

    let (tx, mut rx) = mpsc::channel::<Action>(5000);

    runtime.block_on(async {
        tokio::spawn(manage_interruption(pairs.clone(), config.clone()));

        let retry_layer = RetryBackoffLayer::new(50, 500, 100);
        let client = RpcClient::builder()
            .layer(retry_layer)
            .ws(ws)
            .await
            .map_err(|e| miette!(e))?;
        let provider = Arc::new(ProviderBuilder::new().on_client(client));
        let signer: PrivateKeySigner = "".parse().unwrap();

        info!("Subscribing to the events...");
        tokio::spawn(subscribe(provider.clone(), pairs.clone(), tx.clone()));

        info!("Processing known pairs...");
        process_known_pairs(pairs.clone(), provider.clone())
            .await
            .map_err(|e| miette!(e))?;
        info!("Finished processing known pairs...");

        let mut queue_last_time_not_empty = Instant::now();
        let mut block_processed = false;
        let mut involved_pairs: HashSet<Address> = HashSet::new();

        loop {
            let action = rx.try_recv();

            if let Ok(action) = action {
                queue_last_time_not_empty = Instant::now();
                block_processed = false;
                let len = rx.len();

                debug!(
                    "Processing action {:?}, {:?} actions left",
                    action, len
                );

                match action {
                    Action::ProcessNewPair(pair_address) => {
                        process_new_pair(pairs.clone(), pair_address, provider.clone())
                            .await
                            .map_err(|e| miette!(e))?;
                        involved_pairs.insert(pair_address);
                    }
                    Action::ProcessOldPair(
                        pair_address,
                        amount0_in,
                        amount1_in,
                        amount0_out,
                        amount1_out,
                    ) => {
                        process_old_pair(
                            pairs.clone(),
                            pair_address,
                            amount0_in,
                            amount1_in,
                            amount0_out,
                            amount1_out,
                        )
                            .map_err(|e| miette!(e))?;
                        involved_pairs.insert(pair_address);
                    }
                }
            } else {
                if !block_processed && Instant::now().duration_since(queue_last_time_not_empty) > Duration::from_millis(50) {
                    info!("The actions queue has been empty for 100ms, we assume the entire block has been processed");
                    info!("Involved pairs: {:?}", involved_pairs);

                    look_for_opportunity(pairs.clone(), &involved_pairs);

                    block_processed = true;
                    involved_pairs.clear();
                };

                std::thread::sleep(Duration::from_millis(10));
            }
        }

        #[allow(unreachable_code)]
        Ok(())
    })
}
